// file: ../../sourcen/firmenagen-shell/informationexchange/FirmenAgentPlanInformation.java

package informationexchange;
import informationexchange.PlanInformation;
import knowledge.*;
import informationexchange.PlanTree;
import java.io.Serializable;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;
import java.math.BigDecimal;
import systemcontrol.Results;
import systemcontrol.QueryResults;
import operators.*;
import exceptions.NoTargetPathException;
import exceptions.NetResultNoClassificationException;
import dataprovider.net.NetResult;

/**
   Subklasse der Shell-Klasse PlanInformation. Methoden verify() und generateTargetPredicates() muessen
   fuer jeden konkreten Agenten in einer eigenen Subklasse ausprogrammiert werden.
   @author Christian Fischbach
   @version 1.0
 */
public class FirmenAgentPlanInformation extends PlanInformation {


    /**
        Konstruktor
     */
    public FirmenAgentPlanInformation(long planID, long searchID, A_Box localABox, T_Box tBox, OperatorDB operatorDB, 
                                    boolean createInitialWorld) 
        throws NoTargetPathException {
        super(planID, searchID, localABox, tBox, operatorDB, createInitialWorld);
    }



    /* ------------------------------------------------ */
    /*                 verify-Methoden                  */
    /* ------------------------------------------------ */

    /**
        Fuer den Firmenagenten sollen genau die NetResults von compage A-Box Instanzen 
        zurueckgegeben werden, fuer die die Zielpraedikate erfuellt sind.
     */
    public void verify() {

        //Bei jedem verify()-Aufruf muessen die Results neu berechnet werden
        goalInstances =  new Results(); 
 
        LogService.logPlan(this, "verify() called.");

        Instance companynameI = findConstraintInstance("companyname");
        Instance productnameI = findConstraintInstance("productname");

        //LogService.logPlan(this, "\nconcept2constraintInstance:"+concept2constraintInstance+
        //                         "\ncompanynameI:"+companynameI);

        //Wurde ein companyname vorgegeben?
        if(companynameI != null) {
            //JA! Verifiziere vom companyname ausgehend
            verifyFromCompanyname(companynameI, productnameI);

        } else {
            //NEIN!
            //Wurde ein Produktname vorgegeben?
            if(productnameI!=null) {
                //JA! Verifiziere vom productname ausgehend
                verifyFromProductname(productnameI);

            } else {
                //NEIN! Dann liegt eine ungueltige Eingabe vor.
                LogService.logPlan(this, "verify(): Weder companyname noch productname vorgegeben. Abbruch.");
            };//else1
        };//else1
        
    }//verify()


    /**
        Sucht die passenden compages ausgehend vom vorgegebenen companyname.
        @author Christian Fischbach
        @version 1.0
     */
    protected void verifyFromCompanyname(Instance companynameI, Instance productnameI) {

        LogService.logPlan(this, "verifyFromCompanyname/2 called.");

        //Log-Ausgaben werden gesammelt ausgegeben.
        String verifyMess;

        //Namen der targetPredicates auf ihren Confidence-Wert mappen
	tPredName2Confidence = new Hashtable(4);
        
        //Hole die konkreten has_compage Rollen zum vorgegebenen companyname
        LinkedList has_compageLL = companynameI.getRoles(tBox.lookupRole("has_compage"));
        
        //Gibt es mind. eine konkrete has_compage Rolle?
        if(has_compageLL==null || has_compageLL.size() == 0) { 
            //NEIN! => Ende.
            LogService.logPlan(this, "verify(): die companyname-Instanz der Eingabe hat keine konkrete "+
                                     "has_compage Rolle! Abbruch.");
            return;
        };
                
        //JA!
        //iteriere ueber die konkreten has_compage-Rollen
        ListIterator has_compageLI = has_compageLL.listIterator();

        //Vorab die keyword-Instanz auslesen
        Instance keywordI = findConstraintInstance("keyword"); 
 
        has_compageIt:
        while(has_compageLI.hasNext()) {
	    
            //Die aktuelle has_compage-Rolle auslesen
	    ConcreteRole has_compageCR = (ConcreteRole) has_compageLI.next();

	    //Hole die compage-Instanz aus der has_compage-Rolle
            Instance compageI = has_compageCR.getRangeInstance();

            verifyMess = "\nverify_from_companyname for compage "+compageI+"\n"; 

            //Trade-Instanz (falls vorhanden) auslesen.
            Instance tradeI = findConstraintInstance("trade");
            try {
                //War trade vorgegeben?
                if(tradeI != null) {
                    //JA! 
		    verifyMess += storeRoleConfidence(has_compageCR);
                    
                } else {
                    //NEIN!
                    //Setze rank auf max. Klassifikationswert der Branchensubkonzepte,
                    //nach denen bisher klassifiziert wurde.
                    Iterator ci =  compageI.getNetEntity().getAllClassificators();
                    double maxTradeClassVal = 0.0;
                    while(ci.hasNext()) {
			String ciname = (String) ci.next();
			//TODO: nur Branchenunterkonzepte beruecksichtigen!!
                        double thisTradeClassVal = compageI.getNetEntity().getClassification((String) ci.next());
			if(thisTradeClassVal > maxTradeClassVal) maxTradeClassVal = thisTradeClassVal;
                    };//while

                    verifyMess += storeConfidence("has_compage", maxTradeClassVal);
                    
                };//else
            }
            catch(NetResultNoClassificationException e){
                LogService.logInfo(this, "verify_from_companyname: NetResultNoClassificationException "+
                                         "=> Abbruch.");
                return;
            }
            
                
            //Wurde ein keyword vorgegeben?
	    if(keywordI!=null) {
		//JA!
		ConcreteRole contains_keywordCR = 
		    localABox.getConcreteRole(tBox.lookupRole("contains_keyword"), compageI, keywordI);

		//Existiert die entsprechende contains_keyword-Rolle?
		if(contains_keywordCR!=null) {
		    //JA!
		    verifyMess += storeRoleConfidence(contains_keywordCR);
		} else {
		    //NEIN!
		    verifyMess += storeConfidence("contains_keyword", 0.0);
		};
		
            };//if


            //Wurde productname vorgegeben?
            if(productnameI==null) {
		//NEIN! Kein productname vorgegeben!
                
                //final rank berechnen                 
                computeFinalRank(compageI, verifyMess);

	    } else {
                //JA! productname vorgegeben!

                //Ueberpruefe, ob mind. eine konkrete Rolle has_productpage existiert, so dass die
                //productpage den productname enthaelt (contains_productname Rolle)
                LinkedList has_productpageLL = companynameI.getRoles(tBox.lookupRole("has_productpage"));

		//Gibt es has_productpage-Rollen zum companyname?
                if(has_productpageLL==null || has_productpageLL.size()==0) {
		    //NEIN! es ex. KEINE has_productpage-Rolle
		    verifyMess += storeConfidence("has_productpage", 0.0);
		    //final rank berechnen                 
                    computeFinalRank(compageI, verifyMess);

                } else {
		    //JA! es ex. has_productpage-Rolle
		    ListIterator has_productpageLI = has_productpageLL.listIterator();
		    
		    while(has_productpageLI.hasNext()) {
			//Aktuelle has_productpage Rolle auslesen
			ConcreteRole has_productpageCR = (ConcreteRole) has_productpageLI.next();

			verifyMess += storeRoleConfidence(has_productpageCR);

			//productpage Instanz aus der Rolle auslesen
			Instance productpageI = has_productpageCR.getRangeInstance();
 
			//contains_productname-Rolle (falls vorhanden) auslesen
			ConcreteRole contains_productnameCR = 
			    localABox.getConcreteRole(tBox.lookupRole("contains_productname"), 
						      productpageI, productnameI);

			//Existiert die contains_productname-Rolle?
			if(contains_productnameCR!=null) {
			    //JA!
			    verifyMess += storeRoleConfidence(contains_productnameCR);
			} else {
			    verifyMess += storeConfidence("contains_productname", 0.0);
			};

			//final rank berechnen                 
                        computeFinalRank(compageI, verifyMess);
			    
                    };
                                      
                };//if-else ex. has_productpage-Rolle 
                        
            };//if-else productname vorgegeben

        };//while has_compage      


    }//verifyFromCompanyname()


    /**
        Sucht die passenden compages ausgehend vom vorgegebenen productname.
        @author Christian Fischbach
        @version 1.0
     */
    protected void verifyFromProductname(Instance productnameI) {

        LogService.logPlan(this, "verifyFromProductname/1 called.");

        //Log-Ausgaben werden gesammelt ausgegeben.
        String verifyMess;

        //Namen der targetPredicates auf ihren Confidence-Wert mappen
	tPredName2Confidence = new Hashtable(4);

        //Hole die konkreten contains_productname Rollen zum vorgegebenen productname
        //(Rolle wird hier invers abgefragt!)
        LinkedList contains_productnameLL = productnameI.getInverseRoles(tBox.lookupRole("contains_productname"));

        //Wenn es keine konkreten contains_productname Rollen gibt, dann Abbruch.
        if(contains_productnameLL==null || contains_productnameLL.size()==0) {
            LogService.logInfo(this, "verify: Es ex. keine konkrete contains_productname Rolle. Abbruch!");
            return;
        };

        Instance keywordI = findConstraintInstance("keyword"); 

        //Ueber die konkreten contains_productname Rollen iterieren
        ListIterator contains_productnameLI = contains_productnameLL.listIterator();
        contains_productnameIt:
        while(contains_productnameLI.hasNext()) {
            //Die aktuelle konkrete contains_productname Rolle auslesen
            ConcreteRole contains_productnameCR = (ConcreteRole) contains_productnameLI.next();
            //Die prodpage-Instanz auslesen
            Instance prodpageI = contains_productnameCR.getRangeInstance();//Invers!

            verifyMess = "\nverify_from_productname for productpageI "+prodpageI;
	    
	    verifyMess += storeRoleConfidence(contains_productnameCR);
            
            //Die konkreten has_productpage Rollen dieser prodpage invers auslesen
            LinkedList has_productpageLL = prodpageI.getInverseRoles(tBox.lookupRole("has_productpage"));

            //Wenn es keine konkreten has_productpage Rollen gibt, dann Abbruch.
            if(has_productpageLL==null || has_productpageLL.size()==0) {
		/*
                //hier nur, wenn auch prodpages in die Ausgabe sollen!!! 
		verifyMess += storeConfidence("has_productpage", 0.0);
                verifyMess += storeConfidence("has_compage", 0.0);
                
                if(keywordI!=null) {
		    verifyMess += storeConfidence("contains_keyword", 0.0);
                };
                
                //final rank berechnen 
                computeFinalRank(prodpageI, verifyMess);
		*/

                continue contains_productnameIt; 
	    };
            
            //Ueber die konkreten has_productpage Rollen iterieren
            ListIterator has_productpageLI = has_productpageLL.listIterator();
	    has_productpageIt:
            while(has_productpageLI.hasNext()) {
                ConcreteRole has_productpageCR = (ConcreteRole) has_productpageLI.next();
		
		verifyMess += storeRoleConfidence(has_productpageCR);

                Instance companynameI = has_productpageCR.getRangeInstance();//Invers!
                
                //Die konkreten has_compage Rollen dieses companyname auslesen
                LinkedList has_compageLL = companynameI.getRoles(tBox.lookupRole("has_compage"));

                //Wenn es keine konkreten has_compage Rollen gibt, dann Abbruch.
                if(has_compageLL==null || has_compageLL.size()==0) {
                    
		    /*
		    //hier nur, wenn auch prodpages in die Ausgabe sollen!!!  
                    verifyMess += storeConfidence("has_compage", 0.0);

                    if(keywordI!=null) {
			verifyMess += storeConfidence("contains_keyword", 0.0);
                    };
                

                    //final rank berechnen
		    computeFinalRank(prodpageI, verifyMess);
                    */

                    continue has_productpageIt; 
                };

               
                //Ueber die konkreten has_compage Rollen iterieren
                ListIterator has_compageLI = has_compageLL.listIterator();
                while(has_compageLI.hasNext()) {
                    ConcreteRole has_compageCR = (ConcreteRole) has_compageLI.next();

		    verifyMess += storeRoleConfidence(has_compageCR);

		    Instance compageI = has_compageCR.getRangeInstance();//Nicht invers!
  
                    //Wurde ein keyword vorgegeben?
		    if(keywordI!=null) {
			//JA!
			ConcreteRole contains_keywordCR = 
			    localABox.getConcreteRole(tBox.lookupRole("contains_keyword"), compageI, keywordI);

			//Existiert die entsprechende contains_keyword-Rolle?
			if(contains_keywordCR!=null) {
			    //JA!
			    verifyMess += storeRoleConfidence(contains_keywordCR);
			} else {
			    //NEIN!
			    verifyMess += storeConfidence("contains_keyword", 0.0);
			};
		
		    };//if
                    

                    //final rank berechnen
                    computeFinalRank(compageI, verifyMess);


                };//while has_compage

            };//while has_productpage


	};//while contains_productname

    }//verifyFromProductname



    /* ----------------------------------- */
    /*     Hilfsmethoden fuer verify       */
    /* ----------------------------------- */

    protected void computeFinalRank(Instance compageI, String logMess) {
    
        double tpl = targetPredicates.length;
	double addRank = 0.0,
	       finalRank = 0.0;

        logMess += "computeFinalRank:\n";
        //die confidence-Werte der Zielpraedikate werden summiert ...
	Enumeration predKeysE = tPredName2Confidence.keys();
        while(predKeysE.hasMoreElements()) {
	    String predKey = (String) predKeysE.nextElement();
	    double predConf = ((Double) tPredName2Confidence.get(predKey)).doubleValue();
	    addRank += predConf;
	    logMess += "adding "+predKey+" confidence "+predConf+" (rank now "+addRank+")\n";
	};

	//... und durch Anzahl der Zielpraedikate geteilt
	if(addRank != 0.0) {
	    finalRank = addRank / tpl;
	    logMess += "dividing with "+tpl+"\n";

	    if(finalRank < 0.01) {
		finalRank = 0.01;
	    }

	    double roundedRank = 
		((new BigDecimal(finalRank)).setScale(2, BigDecimal.ROUND_HALF_UP)).doubleValue(); 

	    logMess += "finalRank is: "+finalRank+", rounded: "+roundedRank+" \n";

            goalInstances.add(new QueryResults(roundedRank, (NetResult) compageI.getNetEntity()));

	} else {
	    logMess += "rank out of range! Result will not be added!\n";
	};

	LogService.logInfo(this, logMess);
 

    }//computeFinalRank


    /**
       Sorgt fuer die dynamische Zuordnung von confidence-Werten zu targetPredicates
       innerhalb der verify-Methoden
     */
    protected String storeConfidence(String predKey, double conf) {
	double value = conf;
	String logMess = "storeConfidence: ";

	//Wenn ein "alter" Eintrag existiert, loesche ihn.
	if(tPredName2Confidence.containsKey(predKey)) {
	    tPredName2Confidence.remove(predKey);
	    logMess += "removing "+predKey+" ";
	};

	//ggf. neuen Eintrag einfuegen
	if(value != 0.0) {
	    tPredName2Confidence.put(predKey, new Double(value));
	    logMess += "adding ("+predKey+", "+conf+")\n";
	} else {
	    logMess += "ERROR: "+predKey+" with confidence 0.0\n";
	};

	return logMess;

    }//storeConfidence

    protected String storeRoleConfidence(ConcreteRole tPredRole) {
	return storeConfidence(tPredRole.getRole().name, tPredRole.getConfidence());
    }//storeRoleConfidence



    /* ---------------------------------- */
    /*   Generierung der Zielpraedikate   */
    /* ---------------------------------- */

    /**
        Generiert die Zielpraedikate fuer den Firmenagenten. 
     */
    protected void generateTargetPredicates() {

        LogService.logPlan(this, "generateTargetPredicates() called.");

        concept2constraintInstance = new Hashtable();

        //die generierten targetPreds
        LinkedList theTargetPreds = new LinkedList();

        //Concepts auslesen
	Concept companynameC = tBox.lookup("companyname");
        Concept prodnameC = tBox.lookup("productname");
        Concept prodpageC    = tBox.lookup("productpage");
        Concept compageC     = tBox.lookup("compage");
        Concept keywordC     = tBox.lookup("keyword");
        Concept tradeC       = tBox.lookup("trade");
        Concept tradekeyC    = tBox.lookup("tradekey");   
        //Concept tradepageC =   (Concept) targetConcepts.getFirst();

        //PredArgs generieren (erst nur mit Konzepten)
        PredArgument companynamePA = new PredArgument(companynameC, "comname");
        PredArgument prodnamePA = new PredArgument(prodnameC, "prodname");
        PredArgument prodpagePA    = new PredArgument(prodpageC, "prodpage");
        PredArgument compagePA     = new PredArgument(compageC, "compage");
        PredArgument keywordPA     = new PredArgument(keywordC, "keyword");
        PredArgument tradePA       = new PredArgument(tradeC, "trade");
        PredArgument tradekeyPA    = new PredArgument(tradekeyC, "tradekey");
 
        //ConstraintInstanzen in die PredArgs schreiben
        ListIterator constraintInstancesIt = constraintInstances.listIterator();

        whileIt:
            while(constraintInstancesIt.hasNext()) {
                Instance i = (Instance) constraintInstancesIt.next();
                ConceptWithConfidence icwc = (ConceptWithConfidence) i.getConcepts().getFirst();
                Concept ic = icwc.getConcept();

                concept2constraintInstance.put(ic.name, i);
             
                if(ic.equals(companynameC)) {
                    companynamePA.setInstance(i);
                    continue whileIt;
                };

                if(ic.equals(prodnameC)) {
                    prodnamePA.setInstance(i);
                    continue whileIt;
                };

                if(ic.equals(keywordC)) {
                    keywordPA.setInstance(i);
                    continue whileIt;
                };

            };//while whileIt  


        //Vorab Predicates erzeugen (ohne sie als targetPredicate zu deklarieren)
	//Predicate deals_inPred             = new Predicate(2, "deals_in");
	//deals_inPred.setPredArg(companynamePA, 0);
	//deals_inPred.setPredArg(prodnamePA, 1);

        Predicate has_compagePred          = new Predicate(2, "has_compage");
        has_compagePred.setPredArg(companynamePA, 0);
        has_compagePred.setPredArg(compagePA, 1);        

        Predicate contains_keywordPred     = new Predicate(2, "contains_keyword");
        contains_keywordPred.setPredArg(compagePA, 0);
        contains_keywordPred.setPredArg(keywordPA, 1);

        Predicate contains_prodnamePred = new Predicate(2, "contains_productname");        
        contains_prodnamePred.setPredArg(prodpagePA, 0); 
        contains_prodnamePred.setPredArg(prodnamePA, 1);

        Predicate has_prodpagePred         = new Predicate(2, "has_productpage");
        has_prodpagePred.setPredArg(companynamePA, 0);
        has_prodpagePred.setPredArg(prodpagePA, 1);


        //Predicates zu targetRoles werden targetPredicates
        ListIterator targetRolesIt = targetRoles.listIterator();

        whileIt:
            while(targetRolesIt.hasNext()) {
                Role r = (Role) targetRolesIt.next();
                String n = r.name;
                 
                if(n.equals("has_compage")) {
                    theTargetPreds.add(has_compagePred);
                    continue whileIt;
                };

               if(n.equals("contains_keyword")) {
                    theTargetPreds.add(contains_keywordPred);
                    continue whileIt;
                };      

                if(n.equals("contains_productname")) {
                    theTargetPreds.add(contains_prodnamePred);
                    continue whileIt;
                };

                if(n.equals("has_productpage")) {
                    theTargetPreds.add(has_prodpagePred);
                    continue whileIt;
                };
                
            };//whileIt

        //LinkedList -> Predicates[]
        targetPredicates = (Predicate[]) theTargetPreds.toArray(new Predicate[theTargetPreds.size()]);
       
    }//generateTargetPredicates


    public Instance findConstraintInstance(String conceptName) {

	if(concept2constraintInstance.containsKey(conceptName)) {
            return (Instance) concept2constraintInstance.get(conceptName);
        } else {
            return null;
        };

    }//findConstraintInstance

   
}//class FirmenAgentPlanInformation









