package operators;

import dataanalysis.tools.StringMatcher;
import dataprovider.enumerationobjects.*;
import dataprovider.net.*;
import dataprovider.search.*;
import exceptions.*;
import informationexchange.LogService;
import knowledge.*;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;


/**
   automatisch erzeugter OperatorFindContainsProductOperator
<code>
#------------------------------------------------------------------
# (a) Suche prodpage mit einem search service. In die Suchanfrage
#     geht ein productname & ein trade-synonym ein
# (b) Verifiziere die Ergebnisse nach prodpage mit Klassifikator
# (c) schaue nach, ob das Ergebnis auch tatsaechlich das Produkt
#     enthaelt.
#__________________________________________________________________
# Fuer jeden SearchService gibt es eine Subklasse dieses Operators.
#------------------------------------------------------------------
@name FindContainsProductOperator
@precondition relates_to(syn,trade)
@precondition productname(pn)
@postcondition
@addlistitem contains_productname(prodpage,pn):y,n
</code>
   @author Nils Malzahn
   @version 1.0
*/
public class FindContainsProductOperator extends Operator{
    /**
       Die Suchmaschine die benutzt werden soll.
       @since 1.0
       
    */
    public transient SearchService searchservice;
    
    /**
       Eval-Methode des Operators.
       Naeheres siehe {@link operators.Operator#eval()}.
    */
    public void eval() throws OperatorFailedException{
	final Instance prodname=this.precond[1].getPredArg(0).getInstance();
	final Instance syn=this.precond[0].getPredArg(0).getInstance();
	final Predicate contains_productpage=((Predicate)this.postconds[0].getAddList().getFirst());
	PredArgument prodpagePredArg=contains_productpage.getPredArg(0);
	final Concept prodpageCon=prodpagePredArg.getConcept();
	
	/* Eine entsprechende Suchmaschinenspezifische Anfrage
	   wird vom searchservice erzeugt */
	String searchquery=prodname.name+" "+syn.name;
	SearchNetResult         results;
	SearchResultEnumeration enumeration;
	SearchServiceResult     result;
	
	Instance aktInst;
	
	StringMatcher matcher=new StringMatcher(prodname.name);
	matcher.setIsCaseSensitive(false);
	matcher.setMustFindAll(true);
	
	LinkedList possibleResult=new LinkedList();
	try
	    {
		results = searchservice.doQuery( searchquery );
		enumeration = results.getResults();
		
		while( enumeration.hasMoreElements() == true )
		    {
			result = (SearchServiceResult)enumeration.nextElement();
			LogService.logOperator( this, "trying to download " + result.getURL() );
			try{
			    
			    NetService ns=NetService.getNetService(result.getURL());
			    NetEntity entity=ns.getNetResult(result.getURL());
			    /* erzeugen der Instanz, die evtl. ein Ergebnis darstellt */
			    aktInst = new Instance(entity,prodpageCon);
			    /* Schauen, ob die geholte Seite ueberhaupt eine Produktpage ist */
			    if (prodpageCon.classify(aktInst)){
				/* Es handelt sich also um eine ProdPage */
				
				if(matcher.matchesTo(aktInst.getNetEntity().getText())>0.0){
				/* Wenn ausserdem das gesuchte Produkt als Wort auf
				   der Seite vorkommt, dann ist es ein Kandidat fuer
				   das Endergebnis */
				    possibleResult.add(aktInst);
				};
				/* wenn das Wort nicht vorkommt, dann ignoriere dieses Result */
			    };
			}
			catch(NetServiceException e1){ 
			    LogService.logOperator(this, e1.toString());
			}
			catch(java.net.MalformedURLException e2){
			    LogService.logOperator(this, e2.toString());
			};
			
			    
			/* Wenn es keine Produkpage war, dann ignoriere dieses Result */
		    };


		/* Sortieren der Instanzen nach den Probability-Werten */
		Object liste[] = possibleResult.toArray();
		if( possibleResult.size() > 1 )
			Arrays.sort(liste, new Comparator(){
				public int compare(Object o1, Object o2){
				    Instance inst1=(Instance) o1;
				    Instance inst2=(Instance) o2;
				    return((new Double(inst1.getConfidence(prodpageCon))).compareTo(new Double(inst2.getConfidence(prodpageCon))));
					
				}
			    });
		
		LogService.logOperator(this,liste.length+" moegliche Ergebnisse");
		
		if( liste.length == 0 ){
		    throw new OperatorFailedException("0 moegliche Ergebnisse");
		}

		/* Die letzte Instanz ist dann die am besten zum Konzept gehoerige. Diese
		 kommt in das PredArg.instance waehrend der Rest nur nach PredArg.otherInstances kommt.*/
		Instance resultInstance=(Instance)liste[ liste.length - 1 ];
		prodpagePredArg.setInstance(resultInstance);

		/* Diese Rolle hat entweder die Konfidenz eins oder Null! 
		   An dieser Stelle ist aber klar, dass es eine Loesung geben wird
		   so dass die Konfidenz auf 1.0 gesetzt werden kann.
		 */
		contains_productpage.setConfidence(1.0);

		//LogService.logOperator(this,"1("+resultInstance.getConfidence(prodpageCon)+"): "+resultInstance.name);


		// Die restlichen Instanzen zu other-Instances hinzufuegen.
		for(int pos=(liste.length-2); pos>=0; pos--){
		    resultInstance = (Instance)liste[pos];
		    //LogService.logOperator(this,(liste.length - pos)+"("+resultInstance.getConfidence(prodpageCon)+"): "+resultInstance.name);
		    prodpagePredArg.addInstance(resultInstance);
		};


	    }
	catch( Exception e )
	    {
		LogService.logOperator(this,e.toString());
		if(e instanceof NullPointerException) e.printStackTrace();
		throw new OperatorFailedException("I've got only one postcondition, but I couldn't fullfill it!");
	    };
	
	/* Das ist bei nur einer Nachbedingung immer richtig */    
	this.postcondFullfilled=0;
    }

    public Object clone()throws CloneNotSupportedException{
	FindContainsProductOperator newOp=(FindContainsProductOperator) super.clone();
	newOp.searchservice=this.searchservice;
	return(newOp);
    }
    
    public FindContainsProductOperator(){
	/* 2 Vorbedingung(en) */
	this.precond=new Predicate[2];
	/* Vorbedindungsprdikat 0 */
	this.precond[0]=new Predicate(2,"relates_to");
	
	PredArgument syn = new PredArgument("syn");
	syn.setCreatesWorldObject(false);
	this.precond[0].setPredArg(syn, 0);
	PredArgument trade = new PredArgument("trade");
	trade.setCreatesWorldObject(false);
	this.precond[0].setPredArg(trade, 1);
	
	/* Vorbedindungsprdikat 1 */
	this.precond[1]=new Predicate(1,"productname");
	
	PredArgument pn = new PredArgument("pn");
	pn.setCreatesWorldObject(false);
	this.precond[1].setPredArg(pn, 0);
	Predicate pred;
	/* 1 Nachbedingung(en) */
	this.postconds=new Postcondition[1];
	
	/* Nachbedingung 0 */
	this.postconds[0]=new Postcondition();
	/* AddList */
	pred=new Predicate(2,"contains_productname");
	PredArgument prodpage = new PredArgument("productpage");
	prodpage.setCreatesWorldObject(true);
	pred.setPredArg(prodpage, 0);
	pred.setPredArg(pn, 1);
	this.postconds[0].addAddListPred(pred);

    /* DeleteList */
  }
}





















